/* A ZPU emulator
 *
 * Based on an original idea by Toby Seckshund.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdbool.h>
#include <assert.h>
#include <termio.h>

#include "zpu_memory.h"
#include "zpu.h"

uint32_t pc;
uint32_t sp;
uint32_t tos;
uint8_t  instruction;
uint32_t cpu;

bool touchedPc = false;
bool decodeMask = false;

uint32_t flip(uint32_t i)
{
    uint32_t t = 0;
    for (int j = 0; j < 32; j++)
    {
        t |= ((i >> j) & 1) << (31 - j);
    }
    return t;
}


void emulate()
{
    sp = sp - 4;
    memoryWriteLong(sp, pc + 1);
    pc = (memoryReadByte(pc) - 32) * VECTORSIZE + VECTORBASE;
    touchedPc = true;
}


#define SYS_READ  4
#define SYS_WRITE 5

void syscall()
{    
    int returnAdd = tos;
    int errNoAdd = memoryReadLong(sp + 4);
    int sysCallId = memoryReadLong(sp + 8);
    int fileNo = memoryReadLong(sp + 12);
    int charIndex = memoryReadLong(sp + 16);
    int stringLength = memoryReadLong(sp + 20);
    switch (sysCallId)
    {
	case SYS_WRITE:
	    for (int i = 0; i < stringLength; i++)
	    {
		putchar(memoryReadByte(charIndex++));
	    }	
	    // Return value via R0 (AKA memory address 0)
	    memoryWriteLong(0, stringLength);
	    break;
	case SYS_READ:
	    for (int i = 0; i < stringLength; i++)
	    {
		memoryWriteByte(charIndex++, tty_getchar());
	    }	
	    // Return value via R0 (AKA memory address 0)
	    memoryWriteLong(0, stringLength);
	    break;
    }    
}


static struct termio savemodes;
static int havemodes = 0;


int tty_break()
{
    struct termio modmodes;
    if(ioctl(fileno(stdin), TCGETA, &savemodes) < 0)
    {
      exit(1);
    }
    havemodes = 1;
    modmodes = savemodes;
    modmodes.c_lflag &= ~ICANON;
    modmodes.c_cc[VMIN] = 1;
    modmodes.c_cc[VTIME] = 0;
    return ioctl(fileno(stdin), TCSETAW, &modmodes);
}


int tty_getchar()
{
    return getchar();
}


int tty_fix()
{
    if(!havemodes)
    {
	return (0);
	return ioctl(fileno(stdin), TCSETAW, &savemodes);
    }
}


void zpu_reset()
{
    int i;
    if (tty_break() != 0)
    {
      return exit(1);
    }
//  tty_fix();
    memoryInitialize();
    pc = 0;
    touchedPc = true;
    sp = 0;
    tos = 0;
}


void zpu_execute()
{
    static int step = 0;
    for (;;)
    {
        touchedPc = false;

        instruction = memoryReadByte(pc);
//#define SINGLE_STEP
#ifdef SINGLE_STEP
	printf ("PC=%08x SP=%08x TOS=%08x OP=%02x DM=%02x debug=%08x\n", pc, sp, tos, instruction, decodeMask, 0);
	memoryDisplayLong(0 - 16*4, 16);
	printf("\n");
	getchar();
	if (step++ == 1000)
	{
	    printf("Done.\n");
	    fflush(0);	
	    exit(0);
	}
#endif
        if ((instruction & 0x80) == ZPU_IM)
        {
            if (decodeMask)
            {
                tos = tos << 7;
                tos = tos | instruction & 0x7f;
            }
            else
	    {
		if (sp != 0)
		    memoryWriteLong(sp, tos);		// Flush tos to stack
		else
		    printf("IM : Stack empty\n");
		tos = instruction << 25;		// Sign extend
		tos = ((int32_t)tos) >> 25;
		sp = sp - 4;
	    }
            decodeMask = true;
        }
        else
        {
            decodeMask = false;
            if ((instruction & 0xF0) == ZPU_ADDSP)
            {
                uint32_t addr;
                addr = instruction & 0x0F;
                addr = sp + addr * 4;
		if (sp != 0)
		    memoryWriteLong(sp, tos);		// Flush tos in case it is read back below
		else
		    printf("ADDSP : Stack empty\n");
                tos = tos + memoryReadLong(addr);
            }
            else if ((instruction & 0xE0) == ZPU_LOADSP)
            {
                uint32_t addr;
                addr = (instruction & 0x1F) ^ 0x10; 
                addr = sp + 4 * addr;
		if (sp != 0)
		    memoryWriteLong(sp, tos);		// Flush tos in case it is read back below
		else
		    printf("LOADSP : Stack empty\n");
                tos = memoryReadLong(addr);
		sp = sp - 4;
            }
            else if ((instruction & 0xE0) == ZPU_STORESP)
            {
                uint32_t addr;
                addr = (instruction & 0x1F) ^ 0x10;
                addr = sp + 4 * addr;
                memoryWriteLong(addr, tos);
		sp = sp + 4;
		tos = memoryReadLong(sp);		// Ensure tos is actually from current top of stack
            }
            else
            {
                switch (instruction)
                {
                    case 0:
                        printf ("Breakpoint\n");
			printf ("PC=%08x SP=%08x  OP=%02x DM=%02x debug=%08x\n", pc, sp, instruction, decodeMask, 0);
			fflush(0);	
                        exit(1);
			break;
                    case ZPU_PUSHPC:
			if (sp != 0)
			    memoryWriteLong(sp, tos);	// Flush tos to stack
			else
			    printf("PUSHPC : Stack empty\n");
			sp = sp - 4;
			tos = pc;
                        break;
                    case ZPU_OR:
                    {
			uint32_t a;
			a = tos;
			sp = sp + 4;
			tos = memoryReadLong(sp);
			tos = tos | a;
                        break;
                    }
                    case ZPU_NOT:
                    {
			tos = ~tos;
                        break;
                    }
                    case ZPU_LOAD:
                    {
			if (sp != 0)
			    memoryWriteLong(sp, tos);	// Flush tos in case it is read back below
			else
			    printf("LOAD : Stack empty\n");
                        tos = memoryReadLong(tos);
                        break;
                    }
                    case ZPU_PUSHSPADD:
                    {
			tos =  (tos * 4) + sp;
                        break;
                    }
                    case ZPU_STORE:
		    {
                        uint32_t a;
			a = tos;
			sp = sp + 4;
			tos = memoryReadLong(sp);
			sp = sp + 4;
                        memoryWriteLong(a, tos);
			tos = memoryReadLong(sp);	// Ensure tos is actually from current top of stack
                        break;
		    }
                    case ZPU_POPPC:
                    {
			pc = tos;
			sp = sp + 4;
			touchedPc = true;
			tos = memoryReadLong(sp);	// Ensure tos is actually from current top of stack
                        break;
                    }
                    case ZPU_POPPCREL:
                    {
                        pc = pc + tos;
			sp = sp + 4;
                        touchedPc = true;
			tos = memoryReadLong(sp);	// Ensure tos is actually from current top of stack
                        break;
                    }
                    case ZPU_FLIP:
                    {
			tos = flip(tos);
                        break;
                    }
                    case ZPU_ADD:
                    {
                        int32_t a;
			a = tos;
			sp = sp + 4;
			tos = memoryReadLong(sp);
			tos = tos + a;
                        break;
                    }
                    case ZPU_SUB:
                    {
                        int32_t a;
			a = tos;
			sp = sp + 4;
			tos = memoryReadLong(sp);
			tos = tos - a;
                        break;
                    }
                    case ZPU_PUSHSP:
                    {
			if (sp != 0)
			    memoryWriteLong(sp, tos);	// Flush tos to stack
			else
			    printf("PUSHSP : Stack empty\n");
			tos = sp;
			sp = sp - 4;
                        break;
                    }
                    case ZPU_POPSP:
                    {
			sp = tos;
			tos = memoryReadLong(sp);	// Ensure tos is actually from current top of stack
                        break;
                    }
                    case ZPU_NOP:
                        break;
                    case ZPU_AND:
                    {
                        uint32_t a;
			a = tos;
			sp = sp + 4;
			tos = memoryReadLong(sp);
			tos = tos & a;
                        break;
                    }
                    case ZPU_XOR:
                    {
                        uint32_t a;
			a = tos;
			sp = sp + 4;
			tos = memoryReadLong(sp);
			tos = tos ^ a;
                        break;
                    }
                    case ZPU_LOADB:
                    {
			if (sp != 0)
			    memoryWriteLong(sp, tos);	// Flush tos in case it is read back below
			else
			    printf("LOADB : Stack empty\n");
                        tos = memoryReadByte(tos);
                        break;
                    }
                    case ZPU_STOREB:
                    {
                        uint32_t a;
			a = tos;
			sp = sp + 4;
			tos = memoryReadLong(sp);
			sp = sp + 4;
                        memoryWriteByte(a, tos);
			tos = memoryReadLong(sp);	// Ensure tos is actually from current top of stack
                        break;
                    }
                    case ZPU_LOADH:
                    {
			printf("LOADH : Untested\n");
			if (sp != 0)
			    memoryWriteLong(sp, tos);	// Flush tos in case it is read back below
			else
			    printf("LOADH : Stack empty\n");
                        tos = memoryReadWord(tos);
                        break;
                    }
                    case ZPU_STOREH:
                    {
			printf("STOREH : Untested\n");
                        uint32_t a;
			a = tos;
			sp = sp + 4;
			tos = memoryReadLong(sp);
			sp = sp + 4;
                        memoryWriteWord(a, tos);
			tos = memoryReadLong(sp);	// Ensure tos is actually from current top of stack
                        break;
                    }
                    case ZPU_LESSTHAN:
                    {
                        int32_t a;
			int32_t t;
			a = tos;
			sp = sp + 4;
			t = memoryReadLong(sp);
                        if (a < t)
                            tos = 1;
                        else
                            tos = 0;
                        break;
                    }
                    case ZPU_LESSTHANOREQUAL:
                    {
                        int32_t a;
			int32_t t;
			a = tos;
			sp = sp + 4;
			t = memoryReadLong(sp);
                        if (a <= t)
                            tos = 1;
                        else
                            tos = 0;
                        break;
                    }
                    case ZPU_ULESSTHAN:
                    {
                        uint32_t a;
			a = tos;
			sp = sp + 4;
			tos = memoryReadLong(sp);
                        if (a < tos)
                            tos = 1;
                        else
                            tos = 0;
                        break;
                    }
                    case ZPU_ULESSTHANOREQUAL:
                    {
                        uint32_t a;
			a = tos;
			sp = sp + 4;
			tos = memoryReadLong(sp);
                        if (a <= tos)
                            tos = 1;
                        else
                            tos = 0;
                        break;
                    }
                    case ZPU_SWAP:
                    {
                        tos = ((tos >> 16) & 0xffff) | (tos << 16);
                        break;
                    }
                    case ZPU_MULT16X16:
                    {
                        uint32_t a;
			a = tos;
			sp = sp + 4;
			tos = memoryReadLong(sp);
                        tos = (tos & 0xffff) * (a & 0xffff);
                        break;
                    }
                    case ZPU_EQBRANCH:
                    {
			uint32_t a;
			a = tos;
			sp = sp + 4;
			tos = memoryReadLong(sp);
			sp = sp + 4;
                        if (tos == 0)
                        {
                            pc = pc + a;
                            touchedPc = true;
                        }
			tos = memoryReadLong(sp);
                        break;
                    }
                    case ZPU_NEQBRANCH:
                    {
			uint32_t a;
			a = tos;
			sp = sp + 4;
			tos = memoryReadLong(sp);
			sp = sp + 4;
                        if (tos != 0)
                        {
                            pc = pc + a;
                            touchedPc = true;
                        }
			tos = memoryReadLong(sp);
                        break;
                    }
                    case ZPU_MULT:
                    {
                        int32_t a;
			a = tos;
			sp = sp + 4;
			tos = memoryReadLong(sp);
                        tos = tos * a;
                        break;
                    }
                    case ZPU_DIV:
                    {
                        int32_t a;
			a = tos;
			sp = sp + 4;
			tos = memoryReadLong(sp);
                        if (tos == 0)
                        {
                            printf("Divide by zero\n");
			    fflush(0);	
                            exit(1);
                        }
                        tos = a / tos;
                        break;
                    }
                    case ZPU_MOD:
                    {
                        int32_t a;
			a = tos;
			sp = sp + 4;
			tos = memoryReadLong(sp);
                        if (tos == 0)
                        {
                            printf("Divide (mod) by zero\n");
			    fflush(0);	
                            exit(1);
                        }
                        tos = a % tos;
                        break;
                    }
                    case ZPU_LSHIFTRIGHT:
                    {
                        uint32_t a;
			a = tos;
			sp = sp + 4;
			a = a & 0x3f;
			tos = memoryReadLong(sp);
			tos = tos >> a;
                        break;
                    }
                    case ZPU_ASHIFTLEFT:
                    {
                        uint32_t a;
			a = tos;
			sp = sp + 4;
			a = a & 0x3f;
			tos = memoryReadLong(sp);
			tos = tos << a;
                        break;
                    }
                    case ZPU_ASHIFTRIGHT:
                    {
			uint32_t a;
			int t;
			a = tos;
			sp = sp + 4;
			a = a & 0x3f;
			t = memoryReadLong(sp);
			tos = t >> a;
                        break;
                    }
                    case ZPU_CALL:
                    {
                        uint32_t a;
			a = tos;
			tos = pc + 1;
			pc = a;
                        touchedPc = true;
                        break;
                    }
                    case ZPU_CALLPCREL:
                    {
                        uint32_t a;
			a = tos;
			tos = pc + 1;
			pc = pc + a;
                        touchedPc = true;
                        break;
                    }
                    case ZPU_EQ:
                    {
                        int32_t a;
			a = tos;
			sp = sp + 4;
			tos = memoryReadLong(sp);
                        if (a == tos)
                            tos = 1;
                        else
                            tos = 0;
                        break;
                    }
                    case ZPU_NEQ:
                    {
                        int32_t a;
			a = tos;
			sp = sp + 4;
			tos = memoryReadLong(sp);
                        if (a != tos)
                            tos = 1;
                        else
                            tos = 0;
                        break;
                    }
                    case ZPU_NEG:
                    {
			tos = -tos;
                        break;
                    }
                    case ZPU_CONFIG:
                    {
			cpu = tos;
			sp = sp + 4;
			tos = memoryReadLong(sp);
                        printf ("CONFIG indicates CPU type is %d\n", cpu);
                        break;
		    }
                    case ZPU_SYSCALL:
			syscall();
                        break;
                    default:
                        printf ("Illegal Instruction\n");
			fflush(0);	
                        exit(1);
                        break;
                }
            }
        }
        if (!touchedPc)
        {
            pc = pc + 1;
            touchedPc = true;
        }
    }
}
